
#include <../kernel/trace/trace_probe.h>

#if (LINUX_VERSION_CODE <= KERNEL_VERSION(3,11,0))
struct trace_probe {
  struct list_head        list;
  struct kretprobe        rp;     /* Use rp.kp for kprobe use */
  unsigned long           nhit;
  unsigned int            flags;  /* For TP_FLAG_* */
  const char              *symbol;        /* symbol name */
  struct ftrace_event_class       class;
  struct ftrace_event_call        call;
  struct ftrace_event_file * __rcu *files;
  ssize_t                 size;           /* trace entry size */
  unsigned int            nr_args;
  struct probe_arg        args[];
};
#define make_data_rloc(len, roffs)	\
	(((u32)(len) << 16) | ((u32)(roffs) & 0xffff))
#define get_rloc_len(dl)		((u32)(dl) >> 16)
#define get_rloc_offs(dl)		((u32)(dl) & 0xffff)
#define convert_rloc_to_loc(dl, offs)	((u32)(dl) + (offs))

static inline void *get_rloc_data(u32 *dl)
{
	return (u8 *)dl + get_rloc_offs(*dl);
}
#endif
enum lttng_kretprobe_type {
	EVENT_ENTRY = 0,
	EVENT_RETURN = 1,
	EVENT_KPROBE = 2,
};

struct lttng_krp {
	struct kretprobe krp;
	struct lttng_event *event[2];	/* ENTRY and RETURN */
	struct kref kref_register;
	struct kref kref_alloc;
};


/* Copyed from kernel/trace/trace_kprobes.c added trick for empty strings*/

/* Sum up total data length for dynamic arraies (strings) */
static inline __kprobes int
lttng__get_data_size(struct trace_probe *tp, struct pt_regs *regs)
{
	int i, ret = 0;
	u32 len;

	for (i = 0; i < tp->nr_args; i++)
		if (unlikely(tp->args[i].fetch_size.fn)) {
			call_fetch(&tp->args[i].fetch_size, regs, &len);
			/*
			 * Trick for the common trace format,
			 * empty strings are needed, for babeltrace parsing.
			 */
			if (len == 0)
				ret += 1;
			ret += len;
		}

	return ret;
}

/* Store the value of each argument */
static inline __kprobes void
lttng_store_trace_args(int ent_size, struct trace_probe *tp,
		       struct pt_regs *regs, u8 *data, int maxlen)
{
	int i;
	u32 end = tp->size;
	u32 *dl;	/* Data (relative) location */

	for (i = 0; i < tp->nr_args; i++) {
		if (unlikely(tp->args[i].fetch_size.fn)) {
			/*
			 * First, we set the relative location and
			 * maximum data length to *dl
			 */
			dl = (u32 *)(data + tp->args[i].offset);
			*dl = make_data_rloc(maxlen, end - tp->args[i].offset);
			/* Then try to fetch string or dynamic array data */
			call_fetch(&tp->args[i].fetch, regs, dl);
			if (*dl == 0) {
				u8 *dst;
				*dl = make_data_rloc(1,
						     end - tp->args[i].offset);
				dst = get_rloc_data(dl);
				dst[0] = 0;
				/* Empty zero terminated string */
			}
			/* Reduce maximum length */
			end += get_rloc_len(*dl);
			maxlen -= get_rloc_len(*dl);

			/* Trick here, convert data_rloc to data_loc */
			*dl = convert_rloc_to_loc(*dl,
				 ent_size + tp->args[i].offset);
		} else
			/* Just fetching data normally */

			call_fetch(&tp->args[i].fetch, regs,
				   data + tp->args[i].offset);
	}
}

/*
 * kfree helper for files
 */
static inline __kprobes void kfreefields(const struct lttng_event_desc *desc)
{
	int i;
	for (i = 0; i < desc->nr_fields; i++)
		if (desc->fields[i].name != NULL)
			kfree(desc->fields[i].name);
	kfree(desc->fields);
}

/*
 * Create event description
 */
static
int lttng_create_kprobe_event(const char *name, struct lttng_event *event,
			      enum lttng_kretprobe_type type, char *argin)
{
	static int (*parse_probe_arg)(char *arg, ssize_t *size,
				      struct probe_arg *parg, bool is_return,
				      bool is_kprobe) = NULL;
	struct lttng_event_field *field;
	struct lttng_event_desc *desc;
	struct probe_arg *parg;
	struct trace_probe *tp;
	int argcount = 0;
	ssize_t size = 0;
	char *suffix = "";
	char *alloc_name;
	int add_fields;
	int name_len;
	char *next;
	char *arg;
	int ret;
	int i;

	if (parse_probe_arg == NULL) {
		parse_probe_arg = (void *)kallsyms_lookup_funcptr(
					"traceprobe_parse_probe_arg");
		if (parse_probe_arg == NULL)
			argin = NULL;
		/*
		 * Disables the argument parsing when the parse_probe_arg
		 * function can not be found.
		 */
	}

	if (argin != 0) {
		next = argin;
		while (42) {
			argcount++;
			next = strchr(next, ' ');
			if (next == NULL)
				break;
			while (*next == ' ')
				next++;
		}
	}

	tp = kzalloc(offsetof(struct trace_probe, args) +
		     sizeof(struct probe_arg) * argcount, GFP_KERNEL);
	if (!tp)
		return -ENOMEM;
	switch (type) {
	case EVENT_ENTRY:
		suffix = "_entry";
		event->u.kretprobe.tp[type] = tp;
		break;
	case EVENT_RETURN:
		suffix = "_return";
		event->u.kretprobe.tp[type] = tp;
		break;
	case EVENT_KPROBE:
		event->u.kprobe.tp = tp;
		break;
	}

	desc = kzalloc(sizeof(*event->desc), GFP_KERNEL);
	if (!desc) {
		ret = -ENOMEM;
		goto error_desc;
	}
	desc->name = kstrdup(name, GFP_KERNEL);
	if (!desc->name) {
		ret = -ENOMEM;
		goto error_str;
	}

	name_len = strlen(name);
	name_len += strlen(suffix);

	alloc_name = kmalloc(name_len + 1, GFP_KERNEL);

	if (!alloc_name) {
		ret = -ENOMEM;
		goto error_str;
	}
	strcpy(alloc_name, name);
	strcat(alloc_name, suffix);
	desc->name = alloc_name;

	/* times 2 because per string we need a dummy field */
	desc->fields = field =
		kzalloc(argcount * 2 * sizeof(struct lttng_event_field), GFP_KERNEL);
	if (!field) {
		ret = -ENOMEM;
		goto error_field;
	}
	add_fields = argcount;
	for (i = 0; i < argcount; i++) {
		parg = &tp->args[i];
		arg = strchr(argin, '=');
		if (arg)
			*arg++ = '\0';
		else
			arg = argin;
		if (!arg[0]) {
			ret = -ENOMEM;
			goto error_parse;
		}
		next = strchr(arg, ' ');
		if (next)
			*next++ = '\0';
		ret = parse_probe_arg(arg, &size, parg, 0, 1);
		if (ret)
			goto error_parse;
		if (!strcmp(parg->type->name, "string")) {
			field[i].name = kzalloc(10, GFP_KERNEL);
			snprintf((char *)field[i].name, 10, "dummy%d", i);
			field[i].type.atype = atype_integer;
			field[i].type.u.basic.integer.size = parg->type->size * CHAR_BIT;
			field[i].type.u.basic.integer.alignment = lttng_alignof(unsigned long) * CHAR_BIT;
			field[i].type.u.basic.integer.signedness = lttng_is_signed_type(unsigned long);
			field[i].type.u.basic.integer.reverse_byte_order = 0;
			field[i].type.u.basic.integer.base = 16;
			field[i].type.u.basic.integer.encoding = lttng_encode_none;
			field[add_fields].type.atype = atype_string;
			field[add_fields].name = kstrdup(argin, GFP_KERNEL);
			if (field[i].name == NULL) {
				ret = -ENOMEM;
				goto error_parse;
			}
			add_fields++;
		} else {
			field[i].name = kstrdup(argin, GFP_KERNEL);
			if (field[i].name == NULL) {
				ret = -ENOMEM;
				goto error_parse;
			}

			field[i].type.atype = atype_integer;
			field[i].type.u.basic.integer.size = parg->type->size * CHAR_BIT;
			field[i].type.u.basic.integer.alignment = lttng_alignof(unsigned long) * CHAR_BIT;
			field[i].type.u.basic.integer.signedness = lttng_is_signed_type(unsigned long);
			field[i].type.u.basic.integer.reverse_byte_order = 0;
			field[i].type.u.basic.integer.base = 16;
			field[i].type.u.basic.integer.encoding = lttng_encode_none;
		}
		argin = next;
	}
	desc->nr_fields = add_fields;
	desc->owner = THIS_MODULE;
	event->desc = desc;
	tp->size = size;
	tp->nr_args = argcount;
	return 0;

error_parse:
	kfreefields(desc);
error_field:
	kfree(desc->name);
error_str:
	kfree(desc);
 error_desc:
	kfree(tp);
	return ret;
}
